from ftw.labels.interfaces import ILabeling
from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelSupport
from ftw.labels.utils import make_sortable
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from plone import api
from zope.annotation.interfaces import IAnnotations
from zope.component import adapts
from zope.interface import implements


ANNOTATION_KEY = 'ftw.labels:labeling'


class Labeling(object):
    implements(ILabeling)
    adapts(ILabelSupport)

    def __init__(self, context):
        self.context = context
        self.jar = ILabelJar(self.context)

    def update(self, label_ids):
        jar_keys = self.jar.storage.keys()
        # removes deselected labels
        for label_id in self.storage.keys():  # use keys to avoid RuntimeError: dictionary changed size during iteration
            if label_id not in jar_keys:
                continue  # do we remove key ??
            label = self.jar.get(label_id)
            if label_id not in label_ids:
                if not label.get('by_user', False):
                    self.storage.pop(label_id)

        # adds selected labels
        for label_id in label_ids:
            if label_id not in jar_keys:
                raise LookupError(
                    'Cannot activate label: '
                    'the label "{0}" is not in the label jar. '
                    'Following labels ids are available: {1}'.format(
                        label_id, ', '.join(jar_keys)))
            if label_id not in self.storage:
                self.storage[label_id] = PersistentList()

    def pers_update(self, label_id, activate):
        user_id = self.user_id()
        if not user_id:
            return False
        if activate:
            if label_id not in self.storage:
                self.storage[label_id] = PersistentList()
            if user_id not in self.storage[label_id]:
                self.storage[label_id].append(user_id)
        else:
            if user_id in self.storage[label_id]:
                self.storage[label_id].remove(user_id)
            if not self.storage[label_id]:
                self.storage.pop(label_id)
        return True

    def active_labels(self):
        # selected labels
        labels = []
        for label_id in self.storage:
            try:
                label = self.jar.get(label_id)
                if label.get('by_user', False):
                    if self.user_id() in self.storage[label_id]:
                        labels.append(label)
                else:
                    labels.append(label)
            except KeyError:
                pass
        return sorted(labels, key=lambda cls: make_sortable(cls['title']))

    def available_labels(self):
        # possible labels, marking active ones
        labels = [[], []]
        for label in self.jar.list():
            if label.get('by_user', False):
                label['active'] = (label['label_id'] in self.storage and
                                   self.user_id() in self.storage[label['label_id']])
                labels[0].append(label)
            else:
                label['active'] = (label.get('label_id') in self.storage)
                labels[1].append(label)
        labels[0].sort(key=lambda cls: make_sortable(cls['title']))
        labels[1].sort(key=lambda cls: make_sortable(cls['title']))
        return labels

    def user_id(default=None):
        """ Return current userid """
        cur_user = api.user.get_current()
        if not cur_user:
            return default
        return cur_user.getId()

    @property
    def storage(self):
        if getattr(self, '_storage', None) is None:
            annotation = IAnnotations(self.context)
            if ANNOTATION_KEY not in annotation:
                annotation[ANNOTATION_KEY] = PersistentMapping()
            self._storage = annotation[ANNOTATION_KEY]
        return self._storage



from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelRoot
from ftw.labels.interfaces import ILabelSupport
from ftw.labels.interfaces import ILabeling
from ftw.labels.labeling import Labeling
from ftw.labels.testing import ADAPTERS_ZCML_LAYER
from ftw.testing import MockTestCase
from plone.app.testing import TEST_USER_ID
from zope.annotation import IAttributeAnnotatable
from zope.component import queryAdapter


def label_titles(labels):
    return [label['title'] for label in labels]


class TestLabeling(MockTestCase):
    layer = ADAPTERS_ZCML_LAYER

    def setUp(self):
        super(TestLabeling, self).setUp()
        Labeling.user_id = lambda x: TEST_USER_ID  # needed to avoid plone.api.portal.get error
        self.root = self.providing_stub([ILabelRoot, IAttributeAnnotatable])
        self.document = self.providing_stub([ILabelSupport,
                                             IAttributeAnnotatable])
        self.set_parent(self.document, self.root)
        self.replay()
        self.jar = ILabelJar(self.root)

    def test_adapter(self):
        self.assertTrue(
            queryAdapter(self.document, ILabeling),
            'The labeling adapter is not registered for ILabeling')

    def test_available_labels(self):
        self.jar.add('Question', '#00FF00', False)
        self.jar.add('Read', 'red', True)
        labeling = ILabeling(self.document)
        self.assertEqual(
            [[{'label_id': 'read',
             'title': 'Read',
             'color': 'red',
             'active': False,
             'by_user': True}],
             [{'label_id': 'question',
             'title': 'Question',
             'color': '#00FF00',
             'active': False,
             'by_user': False}]],
            list(labeling.available_labels()))

    def test_available_labels_empty(self):
        labeling = ILabeling(self.document)
        self.assertEqual([[], []], labeling.available_labels())

    def test_available_label(self):
        self.jar.add('Question', '#00FF00', False)
        self.jar.add('Read', 'red', True)
        labeling = ILabeling(self.document)

        labeling.update(['question'])
        labeling.pers_update('read', True)
        self.assertEqual(
            [[{'label_id': 'read',
             'title': 'Read',
             'color': 'red',
             'active': True,
             'by_user': True}],
             [{'label_id': 'question',
              'title': 'Question',
              'color': '#00FF00',
              'active': True,
              'by_user': False}]],
            list(labeling.available_labels()))

    def test_update__enable_labels(self):
        self.jar.add('Bug', 'red', False)
        self.jar.add('Question', 'green', True)
        self.jar.add('Feature', 'purple', True)

        labeling = ILabeling(self.document)
        self.assertEqual([], labeling.active_labels())

        labeling.update(['bug'])
        labeling.pers_update('feature', True)
        self.assertItemsEqual(['Bug', 'Feature'],
                              label_titles(labeling.active_labels()))

    def test_update__disable_labels(self):
        self.jar.add('Bug', 'red', False)
        self.jar.add('Question', 'green', False)
        self.jar.add('Feature', 'purple', True)

        labeling = ILabeling(self.document)
        labeling.update(['bug', 'question'])
        labeling.pers_update('feature', True)
        self.assertItemsEqual(['Bug', 'Feature', 'Question'],
                              label_titles(labeling.active_labels()))

        labeling.update(['bug'])
        labeling.pers_update('feature', False)
        self.assertItemsEqual(['Bug'],
                              label_titles(labeling.active_labels()))

        labeling.update([])
        self.assertEqual([], labeling.active_labels())

    def test_update_raises_LookupError_when_label_not_in_jar(self):
        self.assertEqual(0, len(self.jar.list()))
        self.jar.add('Question', '', False)
        labeling = ILabeling(self.document)
        with self.assertRaises(LookupError) as cm:
            labeling.update(['something'])

        self.assertEqual(
            'Cannot activate label: the label'
            ' "something" is not in the label jar. '
            'Following labels ids are available: question',
            str(cm.exception))

    def test_active_labels(self):
        self.jar.add('Question', '', False)
        self.jar.add('Bug', '', False)
        self.jar.add('Duplicate', '', True)

        labeling = ILabeling(self.document)
        labeling.update(['bug'])
        labeling.pers_update('duplicate', True)
        self.assertListEqual(
            [{'label_id': 'bug',
              'title': 'Bug',
              'color': '',
              'by_user': False},
             {'label_id': 'duplicate',
              'title': 'Duplicate',
              'color': '',
              'by_user': True}],
            labeling.active_labels())

    def test_active_labels_is_sorted(self):
        self.jar.add('Zeta-0', '', False)
        self.jar.add('zeta-1', '', False)
        self.jar.add('alpha-0', '', False)
        self.jar.add('\xc3\x84lpha-1', '', False)
        self.jar.add('Alpha-2', '', False)

        labeling = ILabeling(self.document)
        labeling.update([
            'zeta-0',
            'zeta-1',
            'alpha-0',
            'alpha-1',
            'alpha-2',
        ])

        self.assertEqual(
            ['alpha-0', '\xc3\x84lpha-1', 'Alpha-2', 'Zeta-0', 'zeta-1'],
            [label.get('title') for label in labeling.active_labels()])

    def test_active_labels_filters_deleted_labels(self):
        self.jar.add('Question', 'blue', False)
        self.jar.add('Bug', 'red', False)

        labeling = ILabeling(self.document)
        labeling.update(['question', 'bug'])

        self.jar.remove('bug')

        self.assertEqual(
            [{'label_id': 'question',
             'title': 'Question',
             'color': 'blue',
             'by_user': False}],
            list(labeling.active_labels()))



from ftw.labels.config import COLORS
from ftw.labels.interfaces import ILabelJarChild
from ftw.labels.interfaces import ILabelJar
from ftw.labels.interfaces import ILabelRoot
from ftw.labels.portlets.assignments import LabelJarAssignment
from ftw.labels.utils import title_by_user
from plone.app.portlets.portlets.base import NullAddForm
from plone.app.portlets.portlets.base import Renderer
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile


class AddForm(NullAddForm):

    def create(self):
        return LabelJarAssignment()


class Renderer(Renderer):
    render = ViewPageTemplateFile('labeljar.pt')

    @property
    def available(self):
        if 'portal_factory' in self.context.absolute_url():
            return False
        if ILabelJarChild.providedBy(self.context):
            if not self.label_root_obj:
                return False
        elif not ILabelRoot.providedBy(self.context):
            return False
        if not self.can_edit and not self.labels:
            return False
        return True

    @property
    def labels(self):
        return ILabelJar(self.label_root_obj).list()

    @property
    def colors(self):
        return COLORS

    @property
    def can_edit(self):
        mtool = getToolByName(self.context, 'portal_membership')
        return mtool.checkPermission(
            'ftw.labels: Manage Labels Jar', self.context)

    @property
    def label_root_obj(self):
        try:
            return ILabelJar(self.context)
        except LookupError:
            return None

    def label_title(self, title, by_user):
        return title_by_user(title, by_user)



from Products.Five.browser import BrowserView
from ftw.labels.interfaces import ILabeling
from z3c.json.interfaces import IJSONWriter
from zope.component import getUtility


class Labeling(BrowserView):

    def update(self):
        """Update activated labels.
        """
        labeling = ILabeling(self.context)
        activate_labels = self.request.form.get('activate_labels', [])
        labeling.update(activate_labels)
        self.context.reindexObject(idxs=['labels'])
        return self._redirect()

    def pers_update(self):
        """Update personal labels.
        """
        labeling = ILabeling(self.context)
        label_id = self.request.form.get('label_id')
        activate = self.request.form.get('active')
        if not label_id or activate not in ['True', 'False']:
            return self._redirect()
        activate = not eval(activate)
        ret = labeling.pers_update(label_id, activate)
        self.context.reindexObject(idxs=['labels'])
        writer = getUtility(IJSONWriter)
        self.request.response.setHeader('content-type', 'application/json')
        return writer.write({'ret': (ret and 'ok' or 'nok'), 'new_status': str(activate)})

    def _redirect(self):
        response = self.request.RESPONSE
        referer = self.request.get('HTTP_REFERER')
        if referer and referer is not 'localhost':
            response.redirect(referer)
        else:
            response.redirect(self.context.absolute_url())



from ftw.labels.interfaces import ILabelSupport
from ftw.labels.interfaces import ILabeling
from plone.indexer.decorator import indexer


@indexer(ILabelSupport)
def labels(obj):
    labeling = ILabeling(obj)
    labels = []
    for label_id in labeling.storage:
        try:
            label = labeling.jar.get(label_id)
            if label.get('by_user', False):
                if len(labeling.storage[label_id]):
                    # if at least one user has selected the label, we add it
                    labels.append(label_id)
                    # add each combination user:label
                    for user_id in labeling.storage[label_id]:
                        labels.append('%s:%s' % (user_id, label_id))
            else:
                labels.append(label_id)
        except KeyError:
            pass
    # store something when no label. Query with 'not' in ZCatalog>=3 will retrieve it.
    if not labels:
        return ['_']
    return labels



from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from plone.app.layout.viewlets.common import ViewletBase
from ftw.labels.interfaces import ILabeling
from ftw.labels.interfaces import ILabelSupport
from ftw.labels.utils import title_by_user
from Products.CMFCore.utils import getToolByName


class LabelingViewlet(ViewletBase):

    index = ViewPageTemplateFile('labeling.pt')

    def __init__(self, context, request, view, manager=None):
        super(LabelingViewlet, self).__init__(context, request, view, manager=None)
        self.mtool = getToolByName(self.context, 'portal_membership')

    @property
    def available(self):
        if not self.available_labels:
            return False
        if 'portal_factory' in self.context.absolute_url():
            return False
        if not ILabelSupport.providedBy(self.context):
            return False
        if not self.available_labels[0] and not self.available_labels[1]:
            return False
        return True

    @property
    def active_labels(self):
        return ILabeling(self.context).active_labels()

    @property
    def available_labels(self):
        return ILabeling(self.context).available_labels()

    @property
    def can_edit(self):
        return self.mtool.checkPermission('ftw.labels: Change Labels', self.context)

    @property
    def can_personal_edit(self):
        return self.mtool.checkPermission('ftw.labels: Change Personal Labels', self.context)

    def label_title(self, title, by_user):
        return title_by_user(title, by_user)
